-------------------------------------------------------------------------------
-- Localization libraries
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS = AceLibrary("Babble-Spell-2.2")

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local TEH_BUTTON_NAME = "SimpleSelfRebuffButton"

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE',
	'PLAYER_UPDATE_RESTING',
	'PLAYER_REGEN_ENABLED',
	'PLAYER_CONTROL_GAINED', 'PLAYER_CONTROL_LOST',
	'SimpleSelfRebuff_CheckActivation'
}

local CATEGORY_TRACKING = L["Tracking"] or "Tracking"
local CATEGORY_MAINHAND = L["Main weapon"] or "Main weapon"
local CATEGORY_OFFHAND  = L["Off-hand weapon"] or "Off-hand weapon"

local SOURCE_SPELL = 'spell'
local SOURCE_ITEM  = 'item'

local TARGET_AURA     = 'aura'
local TARGET_TRACKING = 'tracking'
local TARGET_MAINHAND = 'main-hand'
local TARGET_OFFHAND  = 'off-hand'

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local categories = {}

local buffToCast
local castButton

local active

local db

local sources = {}
local targets = {}

local loadFuncs = {}

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local AceOO = AceLibrary('AceOO-2.0')
local Tablet = AceLibrary('Tablet-2.0')
local Gratuity = AceLibrary('Gratuity-2.0')
local Waterfall = AceLibrary:HasInstance('Waterfall-1.0') and AceLibrary('Waterfall-1.0')
 
local Mixin = AceOO.Mixin
local Class = AceOO.Class
local Classpool = AceOO.Classpool

-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

-- Static early initialization
do
	local self = SimpleSelfRebuff

	self:RegisterDB("SimpleSelfRebuffDB")

	local function SCTDisabled()
		return not (self:IsReminderEnabled() and SCT and SCT:IsActive())
	end
	
	local function getReminderColor(key)
		local c = db.reminder.colors[key]
		return c.r, c.g, c.b
	end

	local function setReminderColor(key, r, g, b)
		local c = db.reminder.colors[key]
		c.r, c.g, c.b = r, g, b
	end
		
	self.options = {	
		type = 'group',
		args = {
			menu = {
				type = 'execute',
				name = L["Menu"],
				desc = L["Menu"],
				func = function() Waterfall:Open('SimpleSelfRebuff') end,
				hidden = not Waterfall,
				wfHidden = true,
				order = 50,
			},
			menuHeader = {
				type = 'header',
				name = L["Global options"],
				order = 51,
				wfHidden = true,					
				hidden = not Waterfall,
			},		
			--
			rebuffThreshold = {
				type = 'range',
				name = L["Rebuff time threshold"],
				desc = L["Any buff with remaining time less than this value will be recast. Select 0 to disable."],
				min = 0,
				max = 180,
				step = 1,
				bigStep = 5,
				get = function() return db.rebuffThreshold end,
				set = function(value)
					db.rebuffThreshold = value
					self:CheckRebuff()
				end,
			},
			reminder = {
				type = 'group',
				name = L["Reminder"],
				desc = L["Configure the reminder."],
				args = {
					enable = {
						type = 'toggle',
						name = L["Enable"],
						desc = L["Enable the reminder."],
						get = "IsReminderEnabled",
						set = "ToggleReminder",
						order = 50,
					},
					--
					colorHeader = {
						type = 'header',
						name = L["Colors"],
						order = 79,
					},
					missingColor = {
						type = 'color',
						name = L["Missing message"],
						desc = L["Set the color of the missing message."],
						get = getReminderColor,
						set = setReminderColor,
						passValue = 'missing',
						disabled = "~IsReminderEnabled",
						order = 80,
					},
					refreshingColor = {
						type = 'color',
						name = L["Refreshing message"],
						desc = L["Set the color of the refreshing message."],
						get = getReminderColor,
						set = setReminderColor,
						passValue = 'refreshing',
						disabled = function() return not self:IsReminderEnabled() or db.rebuffThreshold == 0 end,
						order = 80,
					},
					--
					destHeader = {
						type = 'header',
						name = L["Displays"],
						order = 99,
					},
					chat = {
						type = 'toggle',
						name = L["Chat Frame"],
						desc = L["Print reminder into chat frame."],
						passValue = "chat",
						get = "IsReminderDestinationEnabled",
						set = "ToggleReminderDestination", 
						disabled = "~IsReminderEnabled",
					},
					uierror = {
						type = 'toggle',
						name = L["Blizzard error frame"],
						desc = L["Print reminder into Blizzard error frame."],
						passValue = "uierror",
						get = "IsReminderDestinationEnabled",
						set = "ToggleReminderDestination", 
						disabled = "~IsReminderEnabled",
					},
					sctText1 = {
						type = 'toggle',
						name = L["SCT Text #1"],
						desc = L["Show reminder with SCT text #1."],
						passValue = "sctText1",
						get = "IsReminderDestinationEnabled",
						set = "ToggleReminderDestination", 				
						disabled = SCTDisabled,
					},
					sctText2 = {
						type = 'toggle',
						name = L["SCT Text #2"],
						desc = L["Show reminder with SCT text #2."],
						passValue = "sctText2",
						get = "IsReminderDestinationEnabled",
						set = "ToggleReminderDestination",				
						disabled = SCTDisabled,
					},
					sctMessage = {
						type = 'toggle',
						name = L["SCT Message"],
						desc = L["Show reminder as SCT message."],
						passValue = "sctMessage",
						get = "IsReminderDestinationEnabled",
						set = "ToggleReminderDestination",				
						disabled = SCTDisabled,
					},
				},
			},					
			--
			buffHeader = {
				type = 'header',
				name = L["Buff selection"],
				order = 199,
				wfHidden = true,
			},
			--
			fubar = {
				type = 'group',
				name = L["FuBar settings"],
				desc = L["FuBar settings"],
				args = {},
				order = -4,
			}
		}
	}
		
	self.CATEGORY_TRACKING = CATEGORY_TRACKING
	self.CATEGORY_MAINHAND = CATEGORY_MAINHAND 
	self.CATEGORY_OFFHAND  = CATEGORY_OFFHAND 
	
	self.SOURCE_SPELL = SOURCE_SPELL
	self.SOURCE_ITEM  = SOURCE_ITEM
	
	self.TARGET_AURA     = TARGET_AURA
	self.TARGET_TRACKING = TARGET_TRACKING 
	self.TARGET_MAINHAND = TARGET_MAINHAND
	self.TARGET_OFFHAND  = TARGET_OFFHAND

	-- For debugging
	SSR = self
	
	self.categories = categories
	self.sources = sources
	self.targets = targets
end

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 0 then
		msg = msg:format(...)
	end
	return _G.error(msg, 2)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

local function loadHash(t, ...)
	for i = 1, select('#', ...), 2 do
		local k,v = select(i, ...)
		if t[k] == nil then
			t[k] = v
		end
	end
end

local function colorize(text, r, g, b)
	if type(r) == 'table' and r.r then
		return colorize(msg, r.r, r.g, r.b)
	else
		return ("|cff%02x%02x%02x%s|r"):format(math.floor(r*255), math.floor(g*255), math.floor(b*255), text)
	end
end

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == 'table' then
			setmetatable(t, nil)
			for k in pairs(t) do
				t[k] = nil
			end
			list[t] = true
		end
		return nil
	end
end

local function formatDuration(duration, default)
	if not duration or duration < 1 then
		return default or "-"
	elseif duration >= 3600 then
		return HOUR_ONELETTER_ABBR:format(math.ceil(duration/3600))
	elseif duration >= 60 then
		return MINUTE_ONELETTER_ABBR:format(math.ceil(duration/60))
	else
		return SECOND_ONELETTER_ABBR:format(duration)
	end
end

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

function SimpleSelfRebuff:_rev(tag)
	self.revision = math.max(self.revision or 0, tonumber(tag.sub(12, -3)))
end

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self:_rev("$Revision: 42957 $")
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-07-06 08:18:10 -0400 (Fri, 06 Jul 2007) $", 8, 17)

	self:RegisterDefaults('profile', {
		categories = {},
		rebuffThreshold = 30,
		reminder = {
			enabled = true,
			destinations = {
				chat = true,
				sctMessage = true,
			},
			colors = {
				missing    = { r=0.9, g=0.2, b=0.2 },
				refreshing = { r=0.9, g=0.6, b=0.2 },
			},
		},
	})
		
	db = self.db.profile

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	local AceDebug = AceLibrary:HasInstance('AceDebug-2.0') and AceLibrary('AceDebug-2.0')
	if AceDebug then
		AceDebug:embed(self)
		local opt = AceDebug:GetAceOptionsDataTable(self)
		self.options.args.debug = opt.debug
		--SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function SimpleSelfRebuff.Debug() end
	end
	
	if Waterfall then
		Waterfall:Register('SimpleSelfRebuff', 
			'aceOptions', self.options,
			'title','SimpleSelfRebuff',
			'colorR',0.8,
			'colorG',0.2,
			'colorB',0.8
		)
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent(activationCheckEvents, 0.5, 'CheckActivation')
	self:RegisterBucketEvent('PLAYER_AURAS_CHANGED', 0)
	
	-- Should be instant, to handle things before combat
	self:RegisterEvent('PLAYER_REGEN_DISABLED', function() self:SetActive(false) end)

	self:CheckActivation()
end

function SimpleSelfRebuff:OnDisable(first)
	self:SetActive(false)
end

function SimpleSelfRebuff:PLAYER_AURAS_CHANGED()
	self:TriggerEvent('SimpleSelfRebuff_CheckActivation')	
	if active then
		self:ScanPlayerBuffs()
	end
end

function SimpleSelfRebuff:AvailableBuffsChanged()
	self:UpdateOptions()
	self:TriggerEvent('SimpleSelfRebuff_CheckActivation')
end

-------------------------------------------------------------------------------
-- Reminder
-------------------------------------------------------------------------------

function SimpleSelfRebuff:IsReminderEnabled()
	return db.reminder.enabled
end

function SimpleSelfRebuff:ToggleReminder(value)
	db.reminder.enabled = value
end

function SimpleSelfRebuff:IsReminderDestinationEnabled(dest)
	return db.reminder.destinations[dest]
end

function SimpleSelfRebuff:ToggleReminderDestination(dest, value)
	db.reminder.destinations[dest] = value and true or nil
end

do
	local reminderDests = {}
	
	function reminderDests:chat(text, color)
		self:Print("|cff%02x%02x%02x%s|r", color.r*255, color.g*255, color.b*255, text)
	end
	
	function reminderDests:uierror(text, color)
		UIErrorsFrame:AddMessage(text, color.r, color.g, color.b)
	end

	function reminderDests:sctText1(text, color)
		if SCT then
			SCT:DisplayText(text, color, false, "event", 1)
		end
	end

	function reminderDests:sctText2(text, color)
		if SCT then
			SCT:DisplayText(text, color, false, "event", 2)
		end
	end

	function reminderDests:sctMessage(text, color)
		if SCT then
			SCT:DisplayMessage(text, color)
		end
	end
	
	function SimpleSelfRebuff:PrintReminder(buff, timeLeft)
		if not db.reminder.enabled or not next(db.reminder.destinations) then
			return
		end
		local text, color
		if (timeLeft or 0) > 0 then
			text = L["%s expires in %d seconds."]:format(buff.name, formatDuration(timeLeft))
			color = db.reminder.colors.refreshing
		else
			text = L["%s is missing."]:format(buff.name)
			color = db.reminder.colors.missing
		end 
		for dest,value in pairs(db.reminder.destinations) do
			if value and reminderDests[dest] then
				reminderDests[dest](self, text, color)
			end
		end
	end
	
end

-------------------------------------------------------------------------------
-- Options handling
-------------------------------------------------------------------------------

do
	local NO_BUFF = L["None"]
	
	local function getSelectCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and buff.name or NO_BUFF
	end
	
	local function getToggleCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and (buff.name == category.name)
	end
	
	local function setCategoryBuff(category, name)
		if name == NO_BUFF then
			name = nil
		elseif name == true then
			name = category.name
		end
		category:SetExpectedBuff(name)
	end
	
	function SimpleSelfRebuff:BuildOptions()
		local opts = self.options.args
		for categoryName, category in pairs(categories) do
			local categoryName, category = categoryName, category
			local option = {
				name = categoryName,
				desc = categoryName,
				passValue = category,
				set = setCategoryBuff,
				order = 200,
			}
			if category.count > 1 then
				option.type = 'text'
				option.get = getSelectCategoryBuff
				option.validate = { NO_BUFF }
			else
				option.type = 'toggle'
				option.get = getToggleCategoryBuff
			end
			opts[categoryName] = option
		end
		self:UpdateOptions()
	end
	
	function SimpleSelfRebuff:UpdateOptions()
		for categoryName, category in pairs(categories) do
			if category.count > 1 then
				local validate = self.options.args[category.name].validate
				erase(validate)
				for name,buff in pairs(category.buffs) do
					if buff:IsCastable() then
						tinsert(validate, buff.name)
					end
				end
				table.sort(validate)				
				tinsert(validate, 1, NO_BUFF)
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Base buff class
-------------------------------------------------------------------------------
local CategoryClass, BuffClass

BuffClass = Class()

function BuffClass.prototype:init(name, category, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to BuffClass:new() should be a string, not %q', type(name))
	end		
	
	self.name = name
	self.category = category
	loadHash(self, ...)
	
	BuffClass.super.prototype.init(self)
	
	local targetMixin = targets[self.target]
	targetMixin.buffs[self.name] = self
	targetMixin.allBuffs[self] = true
	
	local sourceMixin = sources[self.source]
	sourceMixin.buffs[self.name] = self
	sourceMixin.allBuffs[self] = true
end

function BuffClass.prototype:IsCastable()
	if self.found then
		if self.minLevel and UnitLevel('player') < self.minLevel then
			return false
		elseif self.maxLevel and UnitLevel('player') > self.maxLevel then
			return false
		end
		return true
	end
	return false
end

-------------------------------------------------------------------------------
-- Category class
-------------------------------------------------------------------------------

CategoryClass = Class()

function CategoryClass.prototype:init(name, target, source, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to CategoryClass:new() should be a string, not %q', type(name))
	end
	CategoryClass.super.prototype.init(self)
	self.name = name
	self.buffs = {}
	self.count = 0
	self.target = target or TARGET_AURA
	self.source = source or SOURCE_SPELL
	loadHash(self, ...)
end

function CategoryClass.prototype:add(name, ...)

	if self.buffs[name] then
		error("%q already registered in category %q", name, self.name)
	end
	
	local source = self.source
	local target = self.target
	
	if not sources[source] and loadFuncs[source] then
		loadFuncs[source]()
		loadFuncs[source] = nil
	end
	if not sources[source] then
		error("unknown source for %q: %q", name, source)
	end		
	
	if not targets[target] and loadFuncs[target] then
		loadFuncs[target]()
		loadFuncs[target] = nil
	end	
	if not targets[target] then
		error("unknown target for %q: %q", name, target)
	end		
		
	local class = Classpool(BuffClass, sources[source], targets[target])
	local buff = class:new(name, self, ...)

	self.buffs[buff.name] = buff
	self.count = self.count + 1

	return self
end

function CategoryClass.prototype:addMulti(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:add(name)
	end
	return self
end

function CategoryClass.prototype:GetExpectedBuff()
	local name = db.categories[self.name]
	if self.buffs[name] then
		return self.buffs[name]
	else
		db.categories[self.name] = nil
	end
end

function CategoryClass.prototype:SetExpectedBuff(buff)
	local buffName 
	if type(buff) == 'string' then
		buffName = buff
	elseif buff then
		buffName = buff.name
	end
	if buffName and not self.buffs[buffName] then
		error("Unknown buff %q in category %q", buffName, self.name)
	end
	db.categories[self.name] = buffName
	SimpleSelfRebuff:ScanPlayerBuffs()
	SimpleSelfRebuff:CheckRebuff()
end

function CategoryClass.prototype:GetActualBuff()
	local timeLeft = self.actualBuffExpiration and math.floor(self.actualBuffExpiration - GetTime())
	if not timeLeft or timeLeft <= 0 then
		timeLeft = nil
	end
	return self.actualBuff, self.actualBuffExpiration, timeLeft
end

function CategoryClass.prototype:SetActualBuff(buff, expiration)
	if not buff or not expiration or expiration == 0 then
		expiration = nil
	end
	local dirty = buff ~= self.actualBuff or expiration ~= self.actualBuffExpiration
	self.actualBuff = buff
	self.actualBuffExpiration = expiration
	return dirty
end

-------------------------------------------------------------------------------
-- Some mixin helpers
-------------------------------------------------------------------------------

local function BuffMixin(...)
	local mixin = Mixin{...}
	mixin.allBuffs = {}
	mixin.buffs = {}
	return mixin
end

local function BuffSourceMixin(source, ...)
	mixin = BuffMixin("IsUsable", "SetupBuffSource", "source", ...)
	mixin.source = source
	sources[source] = mixin
	return mixin
end

local function BuffTargetMixin(target, ...)
	mixin = BuffMixin("SetupBuffTarget", "target", ...)
	mixin.target = target
	targets[target] = mixin
	return mixin
end

-------------------------------------------------------------------------------
-- Buffs from items
-------------------------------------------------------------------------------

loadFuncs[SOURCE_ITEM] = function()
	SimpleSelfRebuff:Debug("Loading buff-from-item support")
	
	local ItemBuff = BuffSourceMixin(SOURCE_ITEM)
	local missingItems
	
	local function scheduleScan()
		SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_ScanItems', 0.5, ItemBuff.ScanItems, ItemBuff)
	end
	
	local function enableItemScan(flag)
		if flag and not SimpleSelfRebuff:IsEventRegistered('BAG_UPDATE') then
			SimpleSelfRebuff:RegisterEvent('BAG_UPDATE', scheduleScan)
		elseif SimpleSelfRebuff:IsEventRegistered('BAG_UPDATE') then
			SimpleSelfRebuff:UnregisterEvent('BAG_UPDATE')
		end
	end
	
	function ItemBuff:OnInstanceInit(buff)
		if not buff.itemid then
			error("itemid is missing for ItemBuff %q", buff.name)
		end
		if not missingItems then
			missingItems = new()
		end
		missingItems[buff.itemid] = true
	end
	
	function ItemBuff:OnEnable()
		if next(self.allBuffs) and missingItems then
			self:ScanItems()
		end
	end
	
	function ItemBuff:OnDisable()
		enableItemScan(false)
	end
	
	do
		local itemsToCache
		
		local function CacheItem()
			local itemid = itemsToCache and next(itemsToCache)
			if itemid then
				itemsToCache[itemid] = nil
				SimpleSelfRebuff:Debug('Caching %d', itemid)	
				Gratuity:SetHyperlink("item:"..itemid)
			end
			if not next(itemsToCache) then
				SimpleSelfRebuff:CancelScheduledEvent('SimpleSelfRebuff_CacheItems')
				itemsToCache = del(itemsToCache)
				SimpleSelfRebuff:Debug('Caching done, rescanning')	
				ItemBuff:ScanItems()
			end
		end
	
		function ItemBuff:ScanItems()
			if not missingItems then
				enableItemScan(false)
				return
			end
		
			local dirty
		
			for buff in pairs(self.allBuffs) do
				if not buff.itemid then
					error("Missing itemid for %q", buff.name)
				end
				local itemName = GetItemInfo(buff.itemid)
				if not itemName then
					if not itemsToCache then
						itemsToCache = new()
					end
					itemsToCache[buff.itemid] = true
				else
					if not buff.found then
						buff.found = true							
						dirty = true			
					end
					if buff.name ~= itemName then
						buff.name = itemName
						dirty = true			
					end
					if missingItems then
						missingItems[buff.itemid] = nil
					end
				end
			end
			
			if itemsToCache then
				SimpleSelfRebuff:Debug('Starting caching')	
				SimpleSelfRebuff:ScheduleRepeatingEvent('SimpleSelfRebuff_CacheItems', CacheItem, 0.4)
				enableItemScan(false)
			else
				if missingItems and not next(missingItems) then
					missingItems = del(missingItems)
				end
				enableItemScan(missingItems and next(missingItems))					
			end
			
			if dirty then
				for categoryName, category in pairs(categories) do
					local buffs = category.buffs
					for buffName, buff in pairs(buffs) do
						if buffName ~= buff.name then
							SimpleSelfRebuff:Debug('%q => %q', buffName, buff.name)
							buffs[buff.name] = nil
							buffs[buffName] = buff
						end
					end
				end
				
				SimpleSelfRebuff:AvailableBuffsChanged()
			end
		end
	end
	
	function ItemBuff:IsUsable()
		local start, duration = GetItemCooldown(self.itemid)
		if start > 0 and duration > 0 and start + duration > GetTime() then
			return false
		end					
		return (IsUsableItem(self.itemid))
	end
	
	function ItemBuff:SetupBuffSource(castButton)
		SimpleSelfRebuff:Debug('Setup using for item %q', self.name)
		castButton:SetAttribute('type', 'item')
		castButton:SetAttribute('item', self.name)
	end
	
end

-------------------------------------------------------------------------------
-- Buffs from spells
-------------------------------------------------------------------------------

loadFuncs[SOURCE_SPELL] = function()
	SimpleSelfRebuff:Debug('Loading buff-from-spell support')
		
	local SpellBuff = BuffSourceMixin(SOURCE_SPELL)
	
	local function scheduleScan()
		SimpleSelfRebuff:ScheduleEvent('SimpleSelfRebuff_ScanSpells', 0.5, SpellBuff.ScanSpells, SpellBuff)
	end	
	
	function SpellBuff:OnEnable()
		if next(self.allBuffs) then
			SimpleSelfRebuff:RegisterEvent('SPELLS_CHANGED', scheduleScan)
			self:ScanSpells()
		end
	end
	
	function SpellBuff:OnDisable()
		if SimpleSelfRebuff:IsEventRegistered('SPELLS_CHANGED') then
			SimpleSelfRebuff:UnregisterEvent('SPELLS_CHANGED')
		end
	end
	
	function SpellBuff:ScanSpells()
		local dirty = false
		local knownSpells = new()
		
		local i = 1
		while true do
			local name = GetSpellName(i, BOOKTYPE_SPELL)
			if not name then
				break
			end
			knownSpells[name] = true
			i = i + 1
		end
		
		for buff in pairs(self.allBuffs) do
			local found = knownSpells[buff.name]
			if found and not buff.found then
				-- SimpleSelfRebuff:Debug('New spell: %q', buff.name)
				buff.found = true
				dirty = true
			elseif not found and buff.found then
				-- SimpleSelfRebuff:Debug('Old spell: %q', buff.name)
				buff.found = nil
				dirty = true
			end
		end
		
		knownSpells = del(knownSpells)
	
		if dirty then
			SimpleSelfRebuff:AvailableBuffsChanged()
		end	
	end
	
	function SpellBuff:IsUsable()
		local start, duration = GetSpellCooldown(self.name)
		if start > 0 and duration > 0 and start + duration > GetTime() then
			return false
		end					
		return (IsUsableSpell(self.name))
	end
	
	function SpellBuff:SetupBuffSource(castButton)
		SimpleSelfRebuff:Debug('Setup casting for spell %q', self.name)
		castButton:SetAttribute('type', 'spell')
		castButton:SetAttribute('spell', self.name)
	end
	
end

-------------------------------------------------------------------------------
-- Auras
-------------------------------------------------------------------------------

loadFuncs[TARGET_AURA] = function()
	SimpleSelfRebuff:Debug('Loading aura buff support')

	local AuraBuff = BuffTargetMixin(TARGET_AURA)
	
	function AuraBuff:ScanCurrentBuffs(callback)
		for i = 1, 32 do
			-- Only take in account buffs we actually cast
			local buffName, _, _, _, duration, timeLeft = UnitBuff('player', i)	
			if duration then
				local buff = buffName and self.buffs[buffName]
				if buff then
					callback(buff, timeLeft)
				end
			end
		end	
	end
	
	function AuraBuff:SetupBuffTarget(castButton)
		castButton:SetAttribute('unit', 'player')
	end
	
end

-------------------------------------------------------------------------------
-- Tracking
-------------------------------------------------------------------------------

loadFuncs[TARGET_TRACKING] = function()
	SimpleSelfRebuff:Debug('Loading tracking support')
	
	local TrackingBuff = BuffTargetMixin(TARGET_TRACKING)
	TrackingBuff.textures = {}
	
	function TrackingBuff:OnInstanceInit(buff)
		local texture = BS:GetSpellIcon(buff.name)
		if not texture then
			error('Unknown tracking texture for spell %q', buff.name)
		elseif self.textures[texture] then
			error('Tracking texture already set for %q: %q', buff.name, self.textures[texture].name)
		end
		self.textures[texture] = buff
	end
	
	function TrackingBuff:ScanCurrentBuffs(callback)
		local texture = GetTrackingTexture()
		local buff = texture and self.textures[texture]
		if buff then
			callback(buff)
		end
	end
	
	function TrackingBuff:SetupBuffTarget(castButton)
		castButton:SetAttribute('unit', 'player')
	end

end

-------------------------------------------------------------------------------
-- Weapon enchantment buff
-------------------------------------------------------------------------------

-- Thoses are pretty ugly to handle due to poor API
loadFuncs[TARGET_MAINHAND] = function()
	SimpleSelfRebuff:Debug('Loading weapon buff support')
	
	local weaponEnchants = {}
	
	local function ScanWeaponBuffs(self, callback)
		local timeLeft = weaponEnchants[self.target]
		if timeLeft then
			-- Assume it's the buff we want
			local anyBuff = select(2, next(self.buffs))
			local category = anyBuff.category
			local buff = category:GetExpectedBuff()
			if buff then
				callback(buff, timeLeft)
			end
		end
	end
	
	local function SetupWeaponBuff(self, castButton)
		SimpleSelfRebuff:Debug('Setting up for buff %q on %q', self.name, self.slot)
		castButton:SetAttribute('target-slot', self.slot)
	end
	
	local RefreshWeaponBuffs
	do
		local dirty
		
		local function RefreshOneWeaponBuff(target, hasEnchant, expiration)
			if hasEnchant then
				if not weaponEnchants[target] then
					dirty = true
				end
				weaponEnchants[target] = expiration/1000
			elseif not hasEnchant and weaponEnchants[target] then
				weaponEnchants[target] = nil
				dirty = true
			end
		end
		
		function RefreshWeaponBuffs()
			dirty = nil
			RefreshOneWeaponBuff(TARGET_MAINHAND, GetWeaponEnchantInfo())
			RefreshOneWeaponBuff(TARGET_OFFHAND, select(4, GetWeaponEnchantInfo()))
			if dirty then
				SimpleSelfRebuff:ScanPlayerBuffs()
			end
		end
	end
	
	local OnWeaponBuffEnable, OnWeaponBuffDisable
	do
		local registerCount = 0
	
		function OnWeaponBuffEnable(self)
			if registerCount == 0 then
				SimpleSelfRebuff:ScheduleRepeatingEvent('SimpleSelfRebuff_ScanWeaponBuffs', RefreshWeaponBuffs, 0.5)
				RefreshWeaponBuffs()
			end
			registerCount = registerCount + 1
		end
	
		function OnWeaponBuffDisable(self)
			registerCount = registerCount - 1
			if registerCount == 0 then
				SimpleSelfRebuff:CancelScheduledEvent('SimpleSelfRebuff_ScanWeaponBuffs')			
			end
		end
	end
	 
	local function WeaponBuffMixin(target, slotName)
		local slot = GetInventorySlotInfo(slotName)
		if type(slot) ~= 'number' then
			error("Invalid slot name %q => %q", slotName, slot)
		end
		local mixin = BuffTargetMixin(target, "slot")
		mixin.slot = slot
		mixin.OnEnable = OnWeaponBuffEnable
		mixin.OnDisable = OnWeaponBuffDisable		
		mixin.ScanCurrentBuffs = ScanWeaponBuffs
		mixin.SetupBuffTarget = SetupWeaponBuff
		return mixin
	end
	
	WeaponBuffMixin(TARGET_MAINHAND, 'MainHandSlot')
	WeaponBuffMixin(TARGET_OFFHAND, 'SecondaryHandSlot')
	
end

loadFuncs[TARGET_OFFHAND] = loadFuncs[TARGET_MAINHAND]

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local setupFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(setupFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function, not %q", type(func))
		end
	end
	
	function SimpleSelfRebuff:ProcessRegistry()
		if not setupFuncs then
			return
		end
		
		-- Prebuild categories
		self:GetCategory(self.CATEGORY_TRACKING, self.TARGET_TRACKING)
		self:GetCategory(self.CATEGORY_MAINHAND, self.TARGET_MAINHAND)
		self:GetCategory(self.CATEGORY_OFFHAND,  self.TARGET_OFFHAND)
		
		-- Load all funcs
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(setupFuncs) do
			local succes, msg = pcall(func, self, BS, selfClassName, L)
			if not succes then
				geterrorhandler()(("Error in setupFunc #%d: %s"):format(i, msg))
			end
		end
		
		-- Clean up
		setupFuncs = del(setupFuncs)
		loadFuncs = del(loadFuncs)
		
		self:BuildOptions()
	end
end
	
function SimpleSelfRebuff:GetCategory(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to GetCategory should be a string, not %q", type(name))
	end
	if not categories[name] then
		categories[name] = CategoryClass:new(name, ...)
	end
	return categories[name]
end

function SimpleSelfRebuff:AddStandaloneBuff(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to AddStandaloneBuff should be a string, not %q", type(name))
	end
	self:GetCategory(name):add(name, ...)
end

function SimpleSelfRebuff:AddMultiStandaloneBuffs(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:AddStandaloneBuff(name)
	end
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:SetActive(value)
	if value and not active then
		active = true
		self:Debug('Activated')
		
		for name,source in pairs(sources) do
			if type(source.OnEnable) == "function" then
				source:OnEnable()
			end
		end		
		for name,target in pairs(targets) do
			if type(target.OnEnable) == "function" then
				target:OnEnable()
			end
		end		
	
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_CheckRebuff', self.CheckRebuff, 1.5, self)
		self:ScanPlayerBuffs()
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not value and active then
		active = nil
		self:Debug('Disactivated')
		
		erase(actualBuffs, actualBuffExpirations)

		self:CleanupPendingCast()		
		self:CancelScheduledEvent('SimpleSelfRebuff_CheckRebuff')

		for name,source in pairs(sources) do
			if type(source.OnDisable) == "function" then
				source:OnDisable()
			end
		end		
		for name,target in pairs(targets) do
			if type(target.OnDisable) == "function" then
				target:OnDisable()
			end
		end
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

function SimpleSelfRebuff:CheckActivation()
	self:SetActive(
		not (UnitIsDeadOrGhost('player') or InCombatLockdown() or IsResting()
	  or IsMounted() or IsFlying() or IsStealthed())
	)
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	local rebuffThreshold = db.rebuffThreshold > 0 and db.rebuffThreshold or nil
	local newBuffToCast, buffTimeLeft, shortestBuff, shortestBuffDuration
	
	for categoryName,category in pairs(categories) do
		local expectedBuff = category:GetExpectedBuff()
		if expectedBuff and expectedBuff:IsUsable() then			
			local actualBuff, _, timeLeft = category:GetActualBuff()
			
			-- Check if it is missing
			if not actualBuff then
				newBuffToCast = expectedBuff
				--self:Debug('Missing buff: %q', expectedBuff)
				break
				
			-- Check rebuffing threshold
			elseif rebuffThreshold and timeLeft and actualBuff == expectedBuff then
				if not shortestBuffDuration or timeLeft < shortestBuffDuration then
					shortestBuffDuration, shortestBuff = timeLeft, actualBuff
				end
			end
				
		end
	end
	
	if not newBuffToCast and shortestBuff and shortestBuffDuration < rebuffThreshold then
		newBuffToCast = shortestBuff
		buffTimeLeft = shortestBuffDuration
		self:Debug('Buff needs refresh: %q (%q)', newBuffToCast, shortestBuffDuration)
	end

	if newBuffToCast ~= buffToCast then
		
		if newBuffToCast then
			buffToCast = newBuffToCast
			
			if not castButton then
				castButton = CreateFrame("Button", TEH_BUTTON_NAME, UIParent, "SecureActionButtonTemplate")
				castButton:RegisterForClicks("LeftButtonUp")
				castButton:SetPoint("CENTER",0,0)
				castButton:Hide()
			end
			
			-- Reset attributes
			castButton:SetAttribute('type', nil)
			castButton:SetAttribute('spell', nil)			
			castButton:SetAttribute('unit', nil)			
			castButton:SetAttribute('target-slot', nil)			
			
			-- Setup attributes depending on source and target
			buffToCast:SetupBuffSource(castButton)
			buffToCast:SetupBuffTarget(castButton)
			
			-- Binding
			self:Debug('Binding mousewheel')
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELUP", TEH_BUTTON_NAME, "LeftButton")
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELDOWN", TEH_BUTTON_NAME, "LeftButton")	
			
			-- Reminder
			self:ScheduleEvent('SimpleSelfRebuff_PrintReminder', 2, self.PrintReminder, self, buffToCast, buffTimeLeft > 2 and (buffTimeLeft-2))

		else
			self:CleanupPendingCast()
		end
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end

end

function SimpleSelfRebuff:CleanupPendingCast()
	if buffToCast then
		self:Debug('Unbinding mousewheel')
		if self:IsEventScheduled('SimpleSelfRebuff_PrintReminder') then
			self:CancelScheduledEvent('SimpleSelfRebuff_PrintReminder')
		end		
		buffToCast = nil
		if castButton then
			ClearOverrideBindings(castButton)
		end
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

-------------------------------------------------------------------------------
-- Player buff status tracking
-------------------------------------------------------------------------------

do
	local seen
	local dirty
	
	local function seenBuff(buff, timeLeft)
		--SimpleSelfRebuff:Debug('Seen %s %q (%q)', buff.target, buff.name, timeLeft)
		local category = buff.category
		
		-- Handle subcategories: if there is an expected buff and it has a
		-- subcategory, ignore the detected buff if its subcategory mismatchs
		local expected = category:GetExpectedBuff()
		if expected and expected.subcat and expected.subcat ~= buff.subcat then
			return
		end
		
		local timeLeft = timeLeft and math.floor(timeLeft) or 0
		local expiration = (timeLeft > 0) and (GetTime() + timeLeft)
		seen[category] = true
		if category:SetActualBuff(buff, expiration) then
			dirty = true
		end
	end
	
	function SimpleSelfRebuff:ScanPlayerBuffs()
		seen = new()
		dirty = nil
		
		for target, mixin in pairs(targets) do
			-- self:Debug('Scanning %s', mixin.target)
			mixin:ScanCurrentBuffs(seenBuff)
		end
		
		for categoryName, category in pairs(categories) do
			if not seen[category] then
				if category:SetActualBuff(nil, nil) then
					dirty = true
				end
			end
		end
		
		if dirty then
			dirty = nil
			self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
			--self:Debug('Auras changed - checking rebuff')
			self:CheckRebuff()
		end
		
		seen = del(seen)
		
	end
	
end

-------------------------------------------------------------------------------
-- FuBar Plugin
-------------------------------------------------------------------------------

do
	
	local FuBarPlugin = AceLibrary("FuBarPlugin-2.0")
	
	local plugin = AceLibrary("AceAddon-2.0"):new("FuBarPlugin-2.0", "AceEvent-2.0")
	
	plugin.db = SimpleSelfRebuff:AcquireDBNamespace('FuBarPlugin')
	plugin.hasNoColor = true
	plugin.hideWithoutStandby = true
	plugin.OnMenuRequest = SimpleSelfRebuff.options
	
	function plugin:OnInitialize()
		self.name = SimpleSelfRebuff.name
		self.version = SimpleSelfRebuff.version
		self.revision = SimpleSelfRebuff.revision		
		
		self.OnMenuRequest.args.fubar.args = FuBarPlugin:GetAceOptionsDataTable(self)
	end
	
	function plugin:OnEnable()
		self:RegisterBucketEvent('SimpleSelfRebuff_UpdateDisplay', 0.1, 'UpdateDisplay')
	end
	
	local function GetBuffInfo(category)
	
		local expectedBuff = category:GetExpectedBuff()
		if not expectedBuff then
			return
		end
		
		local actualBuff, _, timeLeft = category:GetActualBuff()
		if not actualBuff then
			return expectedBuff.name, L["Missing"], 1, 0, 0
		end
		
		timeLeft = formatDuration(timeLeft, L["n/a"])
		if actualBuff == expectedBuff then
			return actualBuff.name, timeLeft, 0, 1, 0
		else
			return actualBuff.name, timeLeft, 1, 1, 0
		end
		
	end
	
	function plugin:OnTextUpdate()
		if active then
			if buffToCast then
				local name, timeLeft, r, g, b = GetBuffInfo(buffToCast.category)
				self:SetText(colorize(timeLeft, r, g, b))
				self:ScheduleEvent('SimpleSelfRebuff_UpdateDisplay', 1)
				--self:SetIcon(BS:GetSpellIcon(buffToCast))
			else
				self:SetText('SSR')
				--self:SetIcon(false)
			end
		else
			self:SetText('|cff777777SSR|r')
			--self:SetIcon(false)
		end
	end
	
	function plugin:OnTooltipUpdate()
		local cat
		Tablet:SetTitle(self:GetTitle())
		if not active then
			cat = Tablet:AddCategory()
			cat:AddLine('text', L["Currently inactive (stealthed/in combat/dead/mounted/resting)."])
		else
			cat = Tablet:AddCategory("columns", 3)
			for categoryName, category in pairs(categories) do
				if category.count > 0 then
					local name, timeLeft, r, g, b = GetBuffInfo(category)
					if name then
						cat:AddLine(
							'text',  categoryName,
							'text2', name,     'text2R', r, 'text2G', g, 'text2B', b,
							'text3', timeLeft, 'text3R', r, 'text3G', g, 'text3B', b
						)
					end
				end
			end
			self:ScheduleEvent('SimpleSelfRebuff_UpdateDisplay', 1)		
			Tablet:SetHint(L["Green means \"ok\", yellow means \"override wanted buff\", red means \"missing\"."])
		end
	end

end